#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <stdint.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "vm.h"
#include "task.h"
#include "sched.h"

#define TASK_STACK_SIZE		32

/* Machine scheduler is a simple stack of tasks. The next task to execute is
 * the task from the stack top.
 *
 * There are useful functions where a user can add, delete and schedule a new
 * task.
 */

static sched_state_t	sched_state = SCHED_IDLE;
static struct task	*current;
static struct task	*tasks[TASK_STACK_SIZE];
static unsigned int	tasks_used;

void sched_put_task(struct task *task)
{
	if (tasks_used == TASK_STACK_SIZE)
		xerror(1, "tasks stack overflow");

	tasks[tasks_used++] = task;
}

static struct task *sched_get_task()
{
	if (tasks_used == 0)
		return NULL;

	return tasks[tasks_used-1];
}

/* Delete current task and set scheduler to `idle' state. */
void sched_current_del(struct vm *machine)
{
	assert(tasks_used > 0);

	task_get_ctx(machine, current);

	sched_state = SCHED_IDLE;
	current = NULL;

	tasks[--tasks_used] = NULL;
}

/* Get an active current task. */
struct task *sched_get_current()
{
	return current;
}

/* Get a scheduler state. */
sched_state_t sched_get_state()
{
	return sched_state;
}

/* Reschedule a new task. */
void sched_tasks(struct vm *machine)
{
	struct task *new_task;

	return_if_fail(machine != NULL);

	new_task = sched_get_task();
	/* No tasks in the stack, sched_state is `idle' */
	if (new_task == NULL) {
		assert(sched_state == SCHED_IDLE);
		return;
	}

	/* There is a new task. In `idle' state we set a new task context to
	 * our machine, in other cases we switch save old context and set a
	 * new one.  */
	switch (sched_state) {
	case SCHED_IDLE:
		task_set_ctx(machine, new_task);
		break;
	case SCHED_RUNNING_INTR:
	case SCHED_RUNNING_CYCLE:
	case SCHED_RUNNING_BG:
		task_switch_ctx(machine, current, new_task);
		break;
	default:
		xerror(1, "unknown scheduler state");
	}
	
	/* Adjust a scheduler state to current task mode. */
	switch (new_task->mode) {
	case TASK_MODE_BACKGROUND:
		sched_state = SCHED_RUNNING_BG;
		break;
	case TASK_MODE_CYCLE:
		sched_state = SCHED_RUNNING_CYCLE;
		break;
	case TASK_MODE_INTERRUPT:
		sched_state = SCHED_RUNNING_INTR;
		break;
	default:
		xerror(1, "unknown task type");
	}

	current = new_task;
}

